/*
 * rpmb_accessor.cpp
 *
 * Copyright 2019 Sony Imaging Products & Solutions Inc.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 * 
 */

#include <common.h>
#include <fdtdec.h>
#include <mmc.h>
#include "../../product/cipher/hi_unf_cipher.h"
#ifdef CONFIG_RPMB_REED_SOLOMON
#include "../../lib/rsbep/src/rs.h"
#endif

#define RPMB_BLOCK_SIZE                 256
#define RPMB_BLOCK_DATA_SIZE            223

#define RPMB_MAGIC_NUMBER_SIZE          16
#define RPMB_BOOTCMD_SIZE               100
#define RPMB_BOOTARGS_SIZE              446

#define RPMB_ACCESS_KEY_LENGTH          32

#define BOOT_COUNT_THRESHOLD            5

typedef struct RpmbItemType
{
    s32 primary_block_no_;
    s32 secondary_block_no_;
} RpmbItem;

static const RpmbItem RPMB_ITEM_MAGIC_NUMBER        = {  0, -1 };
static const RpmbItem RPMB_ITEM_BOOTARGS            = {  1,  3 };
static const RpmbItem RPMB_ITEM_BOOTCMD             = {  6,  7 };
static const RpmbItem RPMB_ITEM_BOOT_COUNT          = {  9, 10 };
static const RpmbItem RPMB_ITEM_RECOVERY_BOOTARGS   = { 53, 55 };
static const RpmbItem RPMB_ITEM_RECOVERY_BOOTCMD    = { 58, 59 };

static const char DEVICE_TREE_NODE[] = {"/secure_storage"};
static const char DEVICE_TREE_PROPERTY_MAGIC_NUMBER[] = {"magic"};
static const char DEVICE_TREE_PROPERTY_WORKING_KEY[] = {"temporary_key"};
static const char DEVICE_TREE_PROPERTY_AES_IV[] = {"initial_data"};
static const char DEVICE_TREE_PROPERTY_KEY_SOURCE[] = {"key_id"};
static const char DEVICE_TREE_PROPERTY_ACCESS_KEY[] = {"device_key"};

static u8 rpmb_bootargs[RPMB_BOOTARGS_SIZE + 1];
static u8 rpmb_bootcmd[RPMB_BOOTCMD_SIZE + 1];

typedef struct DecryptParameterType
{
    u8 working_key[16];
    u8 aes_cbc_iv[16];
    HI_UNF_CIPHER_KEY_SRC_E key_source_;
} DecryptParameter;

static int decrypt_data(const DecryptParameter* parameter,
                        const u8* crypted_data,
                        u8* decrypted_data,
                        const size_t data_length)
{
    HI_S32 ret;
    HI_HANDLE handle = 0;
    HI_UNF_CIPHER_CTRL_S cipher_ctrl;

    if ((sizeof(parameter->working_key) > sizeof(cipher_ctrl.u32Key)) ||
        (sizeof(parameter->aes_cbc_iv) > sizeof(cipher_ctrl.u32IV))) {
        return -1;
    }

    ret = HI_UNF_CIPHER_Init();
    if (ret != HI_SUCCESS) {
        return -1;
    }

    ret = HI_UNF_CIPHER_CreateHandle(&handle);
    if (ret != HI_SUCCESS) {
        HI_UNF_CIPHER_DeInit();
        return -1;
    }

    memset(&cipher_ctrl, 0, sizeof(HI_UNF_CIPHER_CTRL_S));
    memcpy(cipher_ctrl.u32Key, parameter->working_key, sizeof(parameter->working_key));
    memcpy(cipher_ctrl.u32IV, parameter->aes_cbc_iv, sizeof(parameter->aes_cbc_iv));
    cipher_ctrl.enAlg       = HI_UNF_CIPHER_ALG_AES;
    cipher_ctrl.enWorkMode  = HI_UNF_CIPHER_WORK_MODE_CBC;
    cipher_ctrl.enBitWidth  = HI_UNF_CIPHER_BIT_WIDTH_128BIT;
    cipher_ctrl.enKeyLen    = HI_UNF_CIPHER_KEY_AES_128BIT;
    cipher_ctrl.enKeySrc    = parameter->key_source_;
    cipher_ctrl.stChangeFlags.bit1IV = 1;   //must set for CBC , CFB mode

    ret = HI_UNF_CIPHER_ConfigHandle(handle, &cipher_ctrl);
    if (ret == HI_SUCCESS) {
        ret = HI_UNF_CIPHER_Decrypt(handle, (HI_U32)crypted_data, (HI_U32)decrypted_data, data_length);
    }

    HI_UNF_CIPHER_DestroyHandle(handle);
    HI_UNF_CIPHER_DeInit();

    return (ret == HI_SUCCESS) ? 0 : -1;
}

static int get_rpmb_parameters(const void *blob,
                               DecryptParameter* parameter,
                               u8* crypted_access_key,
                               const size_t crypted_access_key_length)
{
    int ret;

    int node = fdt_path_offset(blob, DEVICE_TREE_NODE);
    if (node < 0) {
        return node;
    }

    ret = fdtdec_get_byte_array(blob,
                                node,
                                DEVICE_TREE_PROPERTY_WORKING_KEY,
                                parameter->working_key,
                                sizeof(parameter->working_key));
    if (ret < 0) {
        return ret;
    }

    ret = fdtdec_get_byte_array(blob,
                                node,
                                DEVICE_TREE_PROPERTY_AES_IV,
                                parameter->aes_cbc_iv,
                                sizeof(parameter->aes_cbc_iv));
    if (ret < 0) {
        return ret;
    }

    ret = fdtdec_get_int(blob, node, DEVICE_TREE_PROPERTY_KEY_SOURCE, HI_UNF_CIPHER_KEY_SRC_BUTT);
    parameter->key_source_ = (HI_UNF_CIPHER_KEY_SRC_E)ret;

    ret = fdtdec_get_byte_array(blob,
                                node,
                                DEVICE_TREE_PROPERTY_ACCESS_KEY,
                                crypted_access_key,
                                crypted_access_key_length);
    if (ret < 0) {
        return ret;
    }

    return 0;
}

static int change_mmc_partition(struct mmc *mmc, const int device, const char part_num)
{
    if (mmc->part_num != part_num) {
        if (mmc_switch_part(device, part_num) != 0) {
            return -1;
        }
        mmc->part_num = part_num;
    }
    return 0;
}

int rpmb_read_block(struct mmc *mmc,
                    u8 rpmb_read_buffer[RPMB_BLOCK_SIZE],
                    const u32 block_no,
                    const u8* access_key)
{
    int cnt = mmc_rpmb_read(mmc, rpmb_read_buffer, block_no, 1, (u8 *)access_key);
    if (cnt != 1) {
        return -1;
    }

#ifdef CONFIG_RPMB_REED_SOLOMON
    int eras_pos[NN-KK];
    int ret = eras_dec_rs(rpmb_read_buffer, eras_pos, 0);
    if (ret < 0) {
        return -1;
    }
#endif
    return 0;
}

int rpmb_read_data(const RpmbItem* item,
                   u8* buffer,
                   const size_t size,
                   const u8* access_key)
{
    const int curr_device = 0;
    struct mmc *mmc = find_mmc_device(curr_device);
    if (mmc == 0) {
        return -1;
    }

    /* Switch to the RPMB partition */
    char part_num = mmc->part_num;
    int ret = change_mmc_partition(mmc, curr_device, MMC_PART_RPMB);
    if (ret < 0) {
        return -1;
    }

    u8 rpmb_read_buffer[RPMB_BLOCK_SIZE];
    u32 block_count = 0;
    size_t remain_size = size;
    u8* dst_buffer = buffer;

    while (remain_size) {
        ret = rpmb_read_block(mmc, rpmb_read_buffer, item->primary_block_no_ + block_count, access_key);
#ifdef CONFIG_RPMB_REED_SOLOMON
        if (ret < 0) {
            if (item->secondary_block_no_ < 0) {
                break;
            }
            ret = rpmb_read_block(mmc, rpmb_read_buffer, item->secondary_block_no_ + block_count, access_key);
        }
#endif
        if (ret < 0) {
            break;
        }

        u32 copy_size = remain_size;
        if (copy_size > RPMB_BLOCK_DATA_SIZE) {
            copy_size = RPMB_BLOCK_DATA_SIZE;
        }
        memcpy(dst_buffer, &rpmb_read_buffer[0], copy_size);

        remain_size -= copy_size;
        dst_buffer += copy_size;
        ++block_count;
    }

    /* Return to original partition */
    change_mmc_partition(mmc, curr_device, part_num);

    return (size - remain_size);
}

int rpmb_write_data(const RpmbItem* item,
                    const u8* buffer,
                    const size_t size,
                    const u8* access_key)
{
    const int curr_device = 0;
    struct mmc *mmc = find_mmc_device(curr_device);
    if (mmc == 0) {
        return -1;
    }

    /* Switch to the RPMB partition */
    char part_num = mmc->part_num;
    int ret = change_mmc_partition(mmc, curr_device, MMC_PART_RPMB);
    if (ret < 0) {
        return -1;
    }

    u8 rpmb_write_buffer[RPMB_BLOCK_SIZE];
    u32 block_count = 0;
    size_t remain_size = size;
    const u8* src_buffer = buffer;

    while (remain_size) {
        u32 copy_size = remain_size;
        if (copy_size > RPMB_BLOCK_DATA_SIZE) {
            copy_size = RPMB_BLOCK_DATA_SIZE;
        }
        memset(&rpmb_write_buffer[0], 0, RPMB_BLOCK_DATA_SIZE);
        memcpy(&rpmb_write_buffer[0], src_buffer, copy_size);

#ifdef CONFIG_RPMB_REED_SOLOMON
        int ret = encode_rs(rpmb_write_buffer, &rpmb_write_buffer[RPMB_BLOCK_DATA_SIZE]);
        if (ret < 0) {
            return -1;
        }

        if (item->secondary_block_no_ >= 0) {
            int cnt = mmc_rpmb_write(mmc,
                                     rpmb_write_buffer,
                                     item->secondary_block_no_ + block_count,
                                     1,
                                     (u8 *)access_key);
            if (cnt != 1) {
                break;
            }
        }
#endif
        int cnt = mmc_rpmb_write(mmc,
                                 rpmb_write_buffer,
                                 item->primary_block_no_ + block_count, 1,
                                 (u8 *)access_key);
        if (cnt != 1) {
            break;
        }

        remain_size -= copy_size;
        src_buffer += copy_size;
        ++block_count;
    }

    /* Return to original partition */
    change_mmc_partition(mmc, curr_device, part_num);

    return (size - remain_size);
}

int get_rpmb_access_key(const void *blob, u8* decrypted_access_key)
{
    DecryptParameter parameter;
    u8 crypted_access_key[RPMB_ACCESS_KEY_LENGTH];
    int ret;

    memset(&parameter, 0, sizeof(parameter));

    ret = get_rpmb_parameters(blob,
                              &parameter,
                              crypted_access_key,
                              RPMB_ACCESS_KEY_LENGTH);
    if (ret < 0) {
        return ret;
    }

    ret = decrypt_data(&parameter,
                       crypted_access_key,
                       decrypted_access_key,
                       RPMB_ACCESS_KEY_LENGTH);
    if (ret < 0) {
        return ret;
    }

    return 0;
}

int check_magic_number(const void *blob, u8* magic_number)
{
    u8 dt_magic[RPMB_MAGIC_NUMBER_SIZE];
    u8 rpmb_magic[RPMB_MAGIC_NUMBER_SIZE];

    memset(dt_magic, 0, sizeof(dt_magic));
    memset(rpmb_magic, 0, sizeof(rpmb_magic));

    int node = fdt_path_offset(blob, DEVICE_TREE_NODE);
    if (node < 0) {
        return node;
    }

    int ret = fdtdec_get_byte_array(blob, node, DEVICE_TREE_PROPERTY_MAGIC_NUMBER, dt_magic, sizeof(dt_magic));
    if (ret < 0) {
        return ret;
    }

    ret = rpmb_read_data(&RPMB_ITEM_MAGIC_NUMBER, rpmb_magic, sizeof(rpmb_magic), magic_number);
    if (ret != RPMB_MAGIC_NUMBER_SIZE) {
        return -1;
    }

    if (memcmp(dt_magic, rpmb_magic, RPMB_MAGIC_NUMBER_SIZE) != 0) {
        return -1;
    }
    return 0;
}

int get_boot_count(u8* decrypted_access_key)
{
    u8 boot_count;
    int ret = rpmb_read_data(&RPMB_ITEM_BOOT_COUNT, &boot_count, sizeof(boot_count), decrypted_access_key);
    if (ret != sizeof(boot_count)) {
        return -1;
    }
    return boot_count;
}

int set_boot_count(u8* decrypted_access_key, int boot_count)
{
    u8 new_count;
    if (boot_count < 0) {
        new_count = 0;
    }
    else if (boot_count > 0xff) {
        new_count = 0xff;
    }
    else {
        new_count = (u8)(boot_count);
    }
    int ret = rpmb_write_data(&RPMB_ITEM_BOOT_COUNT, &new_count, sizeof(new_count), decrypted_access_key);
    if (ret != sizeof(new_count)) {
        return -1;
    }
    return 0;
}

void setup_bootcmd(const void *blob)
{
    u8 decrypted_access_key[RPMB_ACCESS_KEY_LENGTH];
    int ret = -1;

#ifdef CONFIG_DISABLE_ENV_BOOTCMD
    setenv("bootargs", "");
    setenv("bootcmd", "");
#endif

    ret = get_rpmb_access_key(blob, decrypted_access_key);
    if (ret < 0) {
        return;
    }

    ret = check_magic_number(blob, decrypted_access_key);
    if (ret < 0) {
        return;
    }

    int boot_count = get_boot_count(decrypted_access_key);
    if (boot_count < 0) {
        return;
    }

    // increment boot_count and write to RPMB
    ret = set_boot_count(decrypted_access_key, boot_count + 1);
    if (ret < 0) {
        return;
    }

    const RpmbItem* bootargs_address;
    const RpmbItem* bootcmd_address;
    if (boot_count < BOOT_COUNT_THRESHOLD) {
        bootargs_address = &RPMB_ITEM_BOOTARGS;
        bootcmd_address  = &RPMB_ITEM_BOOTCMD;
    }
    else {
        bootargs_address = &RPMB_ITEM_RECOVERY_BOOTARGS;
        bootcmd_address  = &RPMB_ITEM_RECOVERY_BOOTCMD;
    }

    memset(rpmb_bootargs, 0, sizeof(rpmb_bootargs));
    ret = rpmb_read_data(bootargs_address, rpmb_bootargs, RPMB_BOOTARGS_SIZE, decrypted_access_key);
    if (ret != RPMB_BOOTARGS_SIZE) {
        return;
    }

    memset(rpmb_bootcmd, 0, sizeof(rpmb_bootcmd));
    ret = rpmb_read_data(bootcmd_address, rpmb_bootcmd, RPMB_BOOTCMD_SIZE, decrypted_access_key);
    if (ret != RPMB_BOOTCMD_SIZE) {
        return;
    }
    setenv("bootargs", (char *)rpmb_bootargs);
    setenv("bootcmd", (char *)rpmb_bootcmd);
}
